import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';
import 'package:intl/date_symbol_data_local.dart';
import 'package:intl/intl.dart';

class DatetimePickerDialog extends StatefulWidget {
  final DateTime? datetime;
  final void Function(DateTime?)? setDatetime;
  final bool timePicker;
  final bool? onlyMonthAndYear;

  const DatetimePickerDialog({
    Key? key,
    this.datetime,
    this.setDatetime,
    this.timePicker = false,
    this.onlyMonthAndYear = false,
  }) : super(key: key);

  @override
  _DatetimePickerDialogState createState() => _DatetimePickerDialogState();
}

class _DatetimePickerDialogState extends State<DatetimePickerDialog> {
  DateTime? currentDatetime;

  FixedExtentScrollController? hourScrollController;
  FixedExtentScrollController? minuteScrollController;

  TextEditingController? hourTextController;
  TextEditingController? minuteTextController;

  FixedExtentScrollController? yearScrollController;
  FixedExtentScrollController? monthScrollController;
  FixedExtentScrollController? dayScrollController;

  TextEditingController? yearTextController;
  TextEditingController? monthTextController;
  TextEditingController? dayTextController;

  FocusNode? yearTextFocus;
  FocusNode? monthTextFocus;
  FocusNode? dayTextFocus;

  late bool showTextFields;

  String? lastDayText;
  String? lastMonthText;
  String? lastYearText;

  @override
  void initState() {
    super.initState();

    initializeDateFormatting();

    showTextFields = false;

    if (widget.datetime == null)
      currentDatetime = DateTime(DateTime.now().year, DateTime.now().month,
          widget.onlyMonthAndYear! ? 1 : DateTime.now().day, 0, 0, 0, 0, 0);
    else
      currentDatetime = widget.datetime!.add(Duration.zero);

    if (widget.timePicker) {
      hourScrollController =
          FixedExtentScrollController(initialItem: currentDatetime!.hour);
      minuteScrollController =
          FixedExtentScrollController(initialItem: currentDatetime!.minute);

      hourTextController = TextEditingController(
          text: currentDatetime!.hour.toString().padLeft(2, "0"))
        ..addListener(() {
          if (hourTextController!.text == "" ||
              lastDayText == hourTextController!.text.padLeft(5, "0")) return;
          lastDayText = hourTextController!.text.padLeft(5, "0");
          var hour = int.tryParse(hourTextController!.text);
          if (hour != null && hour >= 0) _setDate(hour: min(hour, 23));
        });
      minuteTextController = TextEditingController(
          text: currentDatetime!.minute.toString().padLeft(2, "0"))
        ..addListener(() {
          if (minuteTextController!.text == "" ||
              lastMonthText == minuteTextController!.text.padLeft(5, "0"))
            return;
          lastMonthText = minuteTextController!.text.padLeft(5, "0");
          var minute = int.tryParse(minuteTextController!.text);
          if (minute != null && minute >= 0) _setDate(minute: min(minute, 59));
        });

      dayTextFocus = _buildFocusNode(hourTextController);
      monthTextFocus = _buildFocusNode(minuteTextController);
    } else {
      yearScrollController =
          FixedExtentScrollController(initialItem: currentDatetime!.year - 1950);
      monthScrollController =
          FixedExtentScrollController(initialItem: currentDatetime!.month - 1);
      dayScrollController =
          FixedExtentScrollController(initialItem: currentDatetime!.day - 1);

      dayTextController =
          TextEditingController(text: currentDatetime!.day.toString())
            ..addListener(() {
              if (dayTextController!.text == "" ||
                  lastDayText == dayTextController!.text.padLeft(5, "0")) return;
              lastDayText = dayTextController!.text.padLeft(5, "0");
              var day = int.tryParse(dayTextController!.text);
              if (day != null && day > 0)
                _setDate(
                    day: min(
                        day,
                        _daysInMonth(DateTime(
                            currentDatetime!.year, currentDatetime!.month))));
            });

      monthTextController =
          TextEditingController(text: currentDatetime!.month.toString())
            ..addListener(() {
              if (monthTextController!.text == "" ||
                  lastMonthText == monthTextController!.text.padLeft(5, "0"))
                return;
              lastMonthText = monthTextController!.text.padLeft(5, "0");
              var month = int.tryParse(monthTextController!.text);
              if (month != null && month > 0) _setDate(month: min(month, 12));
            });

      yearTextController =
          TextEditingController(text: currentDatetime!.year.toString())
            ..addListener(() {
              if (yearTextController!.text == "" ||
                  lastYearText == yearTextController!.text.padLeft(5, "0"))
                return;
              lastYearText = yearTextController!.text.padLeft(5, "0");
              var year = int.tryParse(yearTextController!.text);
              if (year != null && year > 0)
                _setDate(year: max(min(year, 2400), 0));
            });

      dayTextFocus = _buildFocusNode(dayTextController);
      monthTextFocus = _buildFocusNode(monthTextController);
      yearTextFocus = _buildFocusNode(yearTextController);
    }
  }

  FocusNode _buildFocusNode(TextEditingController? textController) {
    var focusNode = FocusNode();
    focusNode.addListener(() {
      if (focusNode.hasFocus) {
        textController!.selection = TextSelection(
            baseOffset: 0, extentOffset: textController.text.length);
      }
    });
    return focusNode;
  }

  @override
  Widget build(BuildContext context) {
    return GestureDetector(
      onTap: () {
        FocusScope.of(context).unfocus();
        setState(() => showTextFields = false);
        Future.delayed(Duration(milliseconds: 100), () {
          _setDate(doUpdateAnyway: true);
        });
      },
      child: DefaultDialog(
        title: Column(
          children: <Widget>[
            Text(
              DateFormat(
                      widget.timePicker
                          ? "HH:mm"
                          : (widget.onlyMonthAndYear!
                              ? "MMM yyyy"
                              : "EEE dd. MMM yyy"),
                      "de_DE")
                  .format(currentDatetime!),
              style: TextStyles.textStyleSize(17),
            ),
            SizedBox(
              height: 25,
            ),
            Container(
              height: 180,
              child: DefaultTextStyle(
                style: TextStyles.textStyleSize(30)
                    .copyWith(color: CleanAndroidDesignDefaultColors.accent),
                child: Row(
                  mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                  children:
                      widget.timePicker ? _getTimeWidgets() : _getDateWidgets(),
                ),
              ),
            ),
          ],
        ),
        dismissible: true,
        buttons: <Widget>[
          DefaultDialogButton(
            titel: Text("Abbrechen"),
            onTap: () {},
          ),
          DefaultDialogButton(
            titel: Text("Fertig"),
            onTap: () => widget.setDatetime!(currentDatetime),
          )
        ],
      ),
    );
  }

  List<Widget> _getTimeWidgets() {
    return showTextFields
        ? <Widget>[
            _getTextField(
                controller: hourTextController,
                focusNode: dayTextFocus,
                nextFocusNode: monthTextFocus),
            Padding(
              padding: const EdgeInsets.only(bottom: 30),
              child: Text(
                ":",
                style: TextStyle(fontSize: 35),
              ),
            ),
            _getTextField(
                controller: minuteTextController, focusNode: monthTextFocus),
          ]
        : <Widget>[
            _getHourScrollView(),
            Padding(
              padding: const EdgeInsets.only(bottom: 30),
              child: Text(
                ":",
                style: TextStyle(fontSize: 35),
              ),
            ),
            _getMinuteScrollView(),
          ];
  }

  Widget _getHourScrollView() {
    return _getListWheelScrollView(
        entries: List<String>.generate(
            24, (index) => (index < 10 ? "0" : "") + index.toString()),
        controller: hourScrollController,
        onChange: (hour) => _setDate(hour: hour),
        textFieldFocusNode: dayTextFocus);
  }

  Widget _getMinuteScrollView() {
    return _getListWheelScrollView(
        entries: List<String>.generate(
            60, (index) => (index < 10 ? "0" : "") + index.toString()),
        controller: minuteScrollController,
        onChange: (minute) => _setDate(minute: minute),
        textFieldFocusNode: monthTextFocus);
  }

  List<Widget> _getDateWidgets() {
    return showTextFields
        ? <Widget>[
            widget.onlyMonthAndYear!
                ? Container()
                : _getTextField(
                    controller: dayTextController,
                    focusNode: dayTextFocus,
                    nextFocusNode: monthTextFocus),
            _getTextField(
                controller: monthTextController,
                focusNode: monthTextFocus,
                nextFocusNode: yearTextFocus),
            _getTextField(
                controller: yearTextController, focusNode: yearTextFocus),
          ]
        : <Widget>[
            widget.onlyMonthAndYear! ? Container() : _getDayScrollView(),
            _getMonthScrollView(),
            _getYearScrollView(),
          ];
  }

  Widget _getTextField(
      {TextEditingController? controller,
      FocusNode? focusNode,
      FocusNode? nextFocusNode}) {
    return Expanded(
      child: Container(
        padding: EdgeInsets.only(bottom: 25, left: 12, right: 10),
        child: Material(
            child: TextField(
          textAlign: TextAlign.center,
          focusNode: focusNode,
          onSubmitted: (s) {
            focusNode!.unfocus();
            if (nextFocusNode == null) {
              setState(() => showTextFields = false);
              Future.delayed(Duration(milliseconds: 100), () {
                _setDate(doUpdateAnyway: true);
              });
            } else
              FocusScope.of(context).requestFocus(nextFocusNode);
          },
          keyboardType: TextInputType.datetime,
          textInputAction: nextFocusNode == null
              ? TextInputAction.done
              : TextInputAction.next,
          controller: controller,
          cursorColor: CleanAndroidDesignDefaultColors.accentLight,
          style: TextStyles.textStyleSize(30)
              .copyWith(color: CleanAndroidDesignDefaultColors.accent),
          decoration: InputDecoration(
            border: InputBorder.none,
          ),
        )),
      ),
    );
  }

  Widget _getDayScrollView() {
    var count =
        _daysInMonth(DateTime(currentDatetime!.year, currentDatetime!.month));
    return _getListWheelScrollView(
        entries:
            List<String>.generate(count, (index) => (index + 1).toString()),
        controller: dayScrollController,
        onChange: (day) => _setDate(day: day + 1),
        textFieldFocusNode: dayTextFocus);
  }

  Widget _getMonthScrollView() {
    return _getListWheelScrollView(
        entries: <String>[
          "JAN.",
          "FEB.",
          "MÄRZ",
          "APR.",
          "MAI",
          "JUNI",
          "JULI",
          "AUG.",
          "SEPT.",
          "OKT.",
          "NOV.",
          "DEZ."
        ],
        controller: monthScrollController,
        onChange: (month) => _setDate(
              month: month + 1,
              day: min(currentDatetime!.day,
                  _daysInMonth(DateTime(currentDatetime!.year, month + 1))),
            ),
        textFieldFocusNode: monthTextFocus);
  }

  Widget _getYearScrollView() {
    return _getListWheelScrollView(
        entries:
            List<String>.generate(90, (index) => (1950 + index).toString()),
        controller: yearScrollController,
        onChange: (year) => _setDate(year: year + 1950),
        textFieldFocusNode: yearTextFocus);
  }

  Widget _getListWheelScrollView(
      {required List<String> entries,
      FixedExtentScrollController? controller,
      void Function(int)? onChange,
      FocusNode? textFieldFocusNode}) {
    return Expanded(
      child: GestureDetector(
        onTapUp: (details) {
          if (details.localPosition.dy > 60 && details.localPosition.dy < 120) {
            setState(() => showTextFields = true);
            Future.delayed(Duration(milliseconds: 100), () {
              _setDate(doUpdateAnyway: true);
              FocusScope.of(context).requestFocus(textFieldFocusNode);
            });
          }
        },
        child: ListWheelScrollView.useDelegate(
          controller: controller,
          onSelectedItemChanged: onChange,
          itemExtent: 60,
          diameterRatio: 20,
          overAndUnderCenterOpacity: 0.2,
          physics: FixedExtentScrollPhysics(),
          childDelegate: ListWheelChildLoopingListDelegate(
              children: entries.map((e) => Text(e)).toList()),
        ),
      ),
    );
  }

  void _setDate(
      {int? year,
      int? month,
      int? day,
      int? hour,
      int? minute,
      bool doUpdateAnyway = false}) {
    var newDateTime = DateTime(
        year ?? currentDatetime!.year,
        month ?? currentDatetime!.month,
        day ?? currentDatetime!.day,
        hour ?? currentDatetime!.hour,
        minute ?? currentDatetime!.minute,
        currentDatetime!.second);

    if (!doUpdateAnyway && currentDatetime!.compareTo(newDateTime) == 0) {
      return;
    }

    setState(() => currentDatetime = newDateTime);

    if (showTextFields) {
      if (widget.timePicker) {
        if (hourTextController!.text !=
            currentDatetime!.hour.toString().padLeft(2, "0"))
          hourTextController!.text =
              currentDatetime!.hour.toString().padLeft(2, "0");

        if (minuteTextController!.text !=
            currentDatetime!.minute.toString().padLeft(2, "0"))
          minuteTextController!.text =
              currentDatetime!.minute.toString().padLeft(2, "0");

        if (hourTextController!.selection.extentOffset -
                hourTextController!.selection.baseOffset <
            2)
          hourTextController!.selection = TextSelection(
              baseOffset: hourTextController!.text.length,
              extentOffset: hourTextController!.text.length);

        if (minuteTextController!.selection.extentOffset -
                minuteTextController!.selection.baseOffset <
            2)
          minuteTextController!.selection = TextSelection(
              baseOffset: minuteTextController!.text.length,
              extentOffset: minuteTextController!.text.length);
      } else {
        if (dayTextController!.text !=
            currentDatetime!.day.toString().padLeft(2, "0"))
          dayTextController!.text =
              currentDatetime!.day.toString().padLeft(2, "0");

        if (monthTextController!.text !=
            currentDatetime!.month.toString().padLeft(2, "0"))
          monthTextController!.text =
              currentDatetime!.month.toString().padLeft(2, "0");

        if (yearTextController!.text !=
            currentDatetime!.year.toString().padLeft(4, "0"))
          yearTextController!.text =
              currentDatetime!.year.toString().padLeft(4, "0");

        if (dayTextController!.selection.extentOffset -
                dayTextController!.selection.baseOffset <
            2)
          dayTextController!.selection = TextSelection(
              baseOffset: dayTextController!.text.length,
              extentOffset: dayTextController!.text.length);

        if (monthTextController!.selection.extentOffset -
                monthTextController!.selection.baseOffset <
            2)
          monthTextController!.selection = TextSelection(
              baseOffset: monthTextController!.text.length,
              extentOffset: monthTextController!.text.length);

        if (yearTextController!.selection.extentOffset -
                yearTextController!.selection.baseOffset <
            4)
          yearTextController!.selection = TextSelection(
              baseOffset: yearTextController!.text.length,
              extentOffset: yearTextController!.text.length);
      }
    } else {
      if (widget.timePicker &&
          hourScrollController!.selectedItem % 24 != currentDatetime!.hour % 24)
        hourScrollController!.jumpToItem(currentDatetime!.hour);

      if (widget.timePicker &&
          minuteScrollController!.selectedItem % 60 !=
              currentDatetime!.minute % 60)
        minuteScrollController!.jumpToItem(currentDatetime!.minute);

      if (!widget.timePicker &&
          yearScrollController!.selectedItem != currentDatetime!.year - 1950)
        yearScrollController!.jumpToItem(currentDatetime!.year - 1950);

      var daysInMonth =
          _daysInMonth(DateTime(currentDatetime!.year, currentDatetime!.month));
      if (!widget.timePicker &&
          dayScrollController!.selectedItem % daysInMonth !=
              (currentDatetime!.day - 1) % daysInMonth)
        dayScrollController!.jumpToItem(currentDatetime!.day - 1);

      if (!widget.timePicker &&
          monthScrollController!.selectedItem % 12 !=
              (currentDatetime!.month - 1) % 12)
        monthScrollController!.jumpToItem(currentDatetime!.month - 1);
    }
  }

  int _daysInMonth(DateTime date) {
    var firstDayThisMonth = new DateTime(date.year, date.month, date.day);
    var firstDayNextMonth = new DateTime(firstDayThisMonth.year,
        firstDayThisMonth.month + 1, firstDayThisMonth.day);
    return firstDayNextMonth.difference(firstDayThisMonth).inDays;
  }
}
