import 'dart:async';
import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:cloud_firestore/cloud_firestore.dart';
import 'package:flutter/material.dart';

class AutoSaveUpdateScaffoldWidget<T extends UpdateableDatabaseItem<T>>
    extends StatefulWidget {
  final IUpdateScaffoldWidgetService<T> databaseCollection;
  final String? updateableItemId;
  final T Function()? createNewItem;
  final T Function(DocumentSnapshot snapshot)? createItemFromSnapshot;

  final void Function(
          BuildContext context, AutoSaveUpdateItemFrame<T>? updateItemFrame)?
      beforeSave;
  final Future<void> Function(
          BuildContext context, AutoSaveUpdateItemFrame<T>? updateItemFrame)?
      beforeSaveAsync;
  final void Function(BuildContext context, DocumentReference itemReference)?
      afterInsert;
  final void Function(BuildContext context, T item)? prepareNewItem;

  final void Function(Validator<T>? validator)? initValidator;
  final int timeoutMilliseconds;

  final List<Widget> Function(
      BuildContext context,
      AutoSaveUpdateItemFrame<T>? updateItemFrame,
      AutoSaveController? autoSaveController)? menuButtonsBuilder;
  final List<Widget>? menuButtons;
  final List<Widget> Function(
      BuildContext context,
      AutoSaveUpdateItemFrame<T>? updateItemFrame,
      AutoSaveController? autoSaveController)? appBarButtonsBuilder;
  final List<Widget>? appBarButtons;
  final Widget Function(
      BuildContext context,
      AutoSaveUpdateItemFrame<T>? updateItemFrame,
      AutoSaveController? autoSaveController)? titleBuilder;
  final List<ScaffoldPageWidget<T>> Function(
      BuildContext context,
      AutoSaveUpdateItemFrame<T>? updateItemFrame,
      ReadOnlyValidator<T>? scaffoldValidator,
      AutoSaveController? autoSaveController)? updatePagesBuilder;
  final Widget Function(
      BuildContext context,
      AutoSaveUpdateItemFrame<T>? updateItemFrame,
      ReadOnlyValidator<T>? validator,
      AutoSaveController? autoSaveController)? builder;
  final IHelpWidget? helpWidget;

  const AutoSaveUpdateScaffoldWidget({
    Key? key,
    required this.databaseCollection,
    this.updateableItemId,
    this.createNewItem,
    this.createItemFromSnapshot,
    this.beforeSave,
    this.beforeSaveAsync,
    this.afterInsert,
    this.prepareNewItem,
    this.initValidator,
    this.timeoutMilliseconds = 4000,
    this.menuButtonsBuilder,
    this.menuButtons,
    this.appBarButtonsBuilder,
    this.appBarButtons,
    this.titleBuilder,
    this.updatePagesBuilder,
    this.builder,
    this.helpWidget,
  }) : super(key: key);

  @override
  _AutoSaveUpdateScaffoldWidgetState<T> createState() =>
      _AutoSaveUpdateScaffoldWidgetState<T>();

  static const int saveStateSaved = 0;
  static const int saveStateSaving = 1;
  static const int saveStateNeedSave = 2;
  static const int saveStateNeedSaveInvalid = 3;
}

class _AutoSaveUpdateScaffoldWidgetState<T extends UpdateableDatabaseItem<T>>
    extends State<AutoSaveUpdateScaffoldWidget<T>> {
  AutoSaveUpdateItemFrame<T>? updateItemFrame;

  int? itemOrigin;
  StreamSubscription<T>? streamSubscriptionDocument;
  StreamSubscription<bool>? streamSubscriptionInternet;

  late StreamController<int> saveStateStreamController;

  Validator<T>? validator;
  List<ScaffoldPageWidget<T>>? pages;

  Key? validatePageKey;
  Key? mainKey;

  AutoSaveController? autoSaveController;

  @override
  void initState() {
    super.initState();

    updateItemFrame = AutoSaveUpdateItemFrame(widget.createItemFromSnapshot);
    updateItemFrame!.setNewItem(widget.createNewItem!());

    validatePageKey = null;
    mainKey = Key(Random().nextInt(10000).toString());
    autoSaveController = AutoSaveController(() => _trySave());
    saveStateStreamController = StreamController.broadcast();
    saveStateStreamController.add(AutoSaveUpdateScaffoldWidget.saveStateSaved);

    if (widget.updateableItemId == null) {
      itemOrigin = DefaultUpdateScaffoldWidget.itemOriginNew;
      try {
        widget.prepareNewItem?.call(context, updateItemFrame!.editableItem);
      } catch (e) {}
    } else {
      itemOrigin = DefaultUpdateScaffoldWidget.itemOriginWaiting;

      void Function(T?) snapshotListener = (T? item) {
        if (!mounted || item == null) return;
        updateItemFrame!.setNewItem(item);
        setState(() {});
      };

      try {
        streamSubscriptionDocument = widget.databaseCollection
            .documentStream(widget.updateableItemId)
            .listen(snapshotListener);
      } catch (e) {}
      var getItem = () async {
        try {
          T? item = await widget.databaseCollection
              .document(widget.updateableItemId, source: Source.server);

          if (item == null) {
            item = await widget.databaseCollection
                .document(widget.updateableItemId, source: Source.cache);

            if (item == null) return DefaultUpdateScaffoldWidget.itemOriginNone;

            snapshotListener(item);
            return DefaultUpdateScaffoldWidget.itemOriginCache;
          } else {
            snapshotListener(item);

            return DefaultUpdateScaffoldWidget.itemOriginServer;
          }
        } catch (e) {
          return DefaultUpdateScaffoldWidget.itemOriginNone;
        }
      };

      getItem()
          .timeout(Duration(milliseconds: widget.timeoutMilliseconds),
              onTimeout: () => DefaultUpdateScaffoldWidget.itemOriginNone)
          .then((origin) {
        if (mounted)
          setState(() {
            itemOrigin = origin;
          });
      });
    }

    streamSubscriptionInternet = CleanAndroidDesignSettings.connectionChange
        ?.listen((_) => setState(() {}));

    Future.delayed(Duration(seconds: 1), autoSave);
  }

  @override
  Widget build(BuildContext context) {
    if (validator == null && widget.initValidator != null) {
      validator = Validator<T>();
      widget.initValidator!(validator);
    }
    pages = widget.updatePagesBuilder?.call(
        context, updateItemFrame, validator?.readOnly(), autoSaveController);

    if (itemOrigin == DefaultUpdateScaffoldWidget.itemOriginWaiting)
      return LoadingScaffoldWidget(
        title: Text("Details"),
        showBackArrow: true,
      );

    if (itemOrigin == DefaultUpdateScaffoldWidget.itemOriginNone)
      return DefaultScaffoldWidget(
        title: Text("Offline"),
        showBackArrow: true,
        child: Column(
          mainAxisSize: MainAxisSize.min,
          mainAxisAlignment: MainAxisAlignment.start,
          children: [
            Container(height: 80),
            Icon(
              Icons.wifi_off,
              size: 50,
              color: Colors.grey[600],
            ),
            DescriptionText(
                "Daten konnten nicht geladen werden, da zurzeit keine Internetvebindung besteht"),
          ],
        ),
      );

    return DefaultScaffoldWidget(
      key: mainKey,
      title: widget.titleBuilder!(context, updateItemFrame, autoSaveController),
      showBackArrow: true,
      onWillPop: _onWillPop,
      menuButtons: widget.menuButtons ??
          widget.menuButtonsBuilder
              ?.call(context, updateItemFrame, autoSaveController),
      appBarButtons: _getAppBarButtons(),
      updatePages: pages,
      child: widget.builder?.call(
          context, updateItemFrame, validator?.readOnly(), autoSaveController),
      defaultPageKey: validatePageKey,
      helpWidget: widget.helpWidget,
    );
  }

  late DateTime lastChangeInData;
  T? lastCheckedItem;

  Future<void> autoSave() async {
    if (!mounted) return;

    var dataHasChanged = lastCheckedItem == null ||
        lastCheckedItem!.toMap().keys.any((key) =>
            key != "last_modified" &&
            (lastCheckedItem!.toMap()[key]?.toString() ?? "").compareTo(
                    updateItemFrame!.editableItem.toMap()[key]?.toString() ??
                        "") !=
                0);
    lastCheckedItem = updateItemFrame!.editableItem.copy();
    if (dataHasChanged) lastChangeInData = DateTime.now();

    var needSave = updateItemFrame!.needSave();
    var isValid = _isValid();

    if (needSave) if (isValid)
      saveStateStreamController
          .add(AutoSaveUpdateScaffoldWidget.saveStateNeedSave);
    else
      saveStateStreamController
          .add(AutoSaveUpdateScaffoldWidget.saveStateNeedSaveInvalid);
    else
      saveStateStreamController
          .add(AutoSaveUpdateScaffoldWidget.saveStateSaved);

    var lastChangeSeconds =
        lastChangeInData.difference(DateTime.now()).abs().inSeconds;

    if (needSave &&
        lastChangeSeconds > 5 &&
        (await CleanAndroidDesignSettings.checkIsConnected())) {
      _saveItem();
    }

    var nextAutoSaveIn = 0;
    if (lastChangeSeconds < 20)
      nextAutoSaveIn = 500;
    else if (lastChangeSeconds < (2 * 60))
      nextAutoSaveIn = 3000;
    else
      nextAutoSaveIn = 25000;
    Future.delayed(Duration(milliseconds: nextAutoSaveIn), autoSave);
  }

  List<Widget> _getAppBarButtons() {
    var list = widget.appBarButtons ??
        widget.appBarButtonsBuilder
            ?.call(context, updateItemFrame, autoSaveController) ??
        <Widget>[];
    if (itemOrigin == DefaultUpdateScaffoldWidget.itemOriginCache)
      list.add(IconButton(
        icon: Icon(
          Icons.signal_wifi_off,
        ),
        tooltip: "Offline Daten",
        onPressed: null,
      ));

    if (itemOrigin == DefaultUpdateScaffoldWidget.itemOriginWaiting)
      list.add(IconButton(
        icon: Container(
          width: 18,
          height: 18,
          child: CircularProgressIndicator(
            valueColor: AlwaysStoppedAnimation<Color?>(
                CleanAndroidDesignDefaultColors.defaultColor),
            strokeWidth: 2,
          ),
        ),
        onPressed: null,
        tooltip: "Lade Daten von Server",
      ));
    list.add(StreamBuilder<int>(
      stream: saveStateStreamController.stream,
      builder: (context, saveStateSnapshot) {
        if (saveStateSnapshot.data ==
            AutoSaveUpdateScaffoldWidget.saveStateNeedSave)
          return IconButton(
            icon: Icon(Icons.save),
            onPressed: () => _trySave(),
            tooltip: "Manuell speichern",
          );
        if (saveStateSnapshot.data ==
            AutoSaveUpdateScaffoldWidget.saveStateNeedSaveInvalid)
          return IconButton(
            icon:
                Icon(Icons.save, color: CleanAndroidDesignDefaultColors.error),
            onPressed: () => _trySave(),
            tooltip: "Eingabe fehlerhaft",
          );
        if (saveStateSnapshot.data ==
            AutoSaveUpdateScaffoldWidget.saveStateSaved)
          return IconButton(
            icon: Icon(Icons.save),
            onPressed: null,
            tooltip: "Kein Speichern nötig",
          );
        else
          return IconButton(
            icon: Container(
              width: 18,
              height: 18,
              child: CircularProgressIndicator(
                valueColor: AlwaysStoppedAnimation<Color?>(
                    CleanAndroidDesignDefaultColors.defaultColor),
                strokeWidth: 2,
              ),
            ),
            onPressed: null,
            tooltip: saveStateSnapshot.data ==
                    AutoSaveUpdateScaffoldWidget.saveStateSaving
                ? "Wird gespeichert..."
                : "Lade Daten...",
          );
      },
    ));

    return list;
  }

  void _trySave() async {
    if (updateItemFrame!.needSave()) {
      if (_isValid(jumpToInvalid: true)) {
        if (await CleanAndroidDesignSettings.checkIsConnected()) {
          _saveItem();
        } else {
          saveStateStreamController
              .add(AutoSaveUpdateScaffoldWidget.saveStateNeedSave);
          SnackBarController.showText(
              "Keine Internetverbindung. Speichern fehlgeschlagen");
        }
      } else {
        saveStateStreamController
            .add(AutoSaveUpdateScaffoldWidget.saveStateNeedSaveInvalid);
      }
    } else {
      saveStateStreamController
          .add(AutoSaveUpdateScaffoldWidget.saveStateSaved);
    }
  }

  bool _isValid({bool jumpToInvalid = false}) {
    var result = validator?.isValid(updateItemFrame!.editableItem);
    if (validator == null || result!.isValid) {
      if (validatePageKey != null) setState(() => validatePageKey = null);
      return true;
    } else if (jumpToInvalid) {
      if (pages != null) {
        setState(() {
          validatePageKey = result.firstInvalidPage;
          mainKey = Key(Random().nextInt(10000).toString());
        });
        Future.delayed(Duration(milliseconds: 100), () {
          var result = validator!.isValid(updateItemFrame!.editableItem);
          if (result.firstInvalidNode != null) {
            FocusScope.of(context).unfocus();
            FocusScope.of(context).requestFocus(result.firstInvalidNode);
          }
        });
      } else {
        setState(() {});
        Future.delayed(Duration(milliseconds: 100), () {
          FocusScope.of(context).unfocus();
          FocusScope.of(context).requestFocus(result.firstInvalidNode);
        });
      }
      return false;
    }
    return false;
  }

  Future<bool> _onWillPop() async {
    if (updateItemFrame!.needSave()) {
      if (_isValid(jumpToInvalid: true)) {
        if (await CleanAndroidDesignSettings.checkIsConnected()) {
          _saveItem();
          return true;
        } else {
          saveStateStreamController
              .add(AutoSaveUpdateScaffoldWidget.saveStateNeedSave);
          var dismiss = false;
          await showCustomDialog(
              context,
              DefaultDialog(
                dismissible: false,
                title: Text(
                    "Es besteht keine Internetverbindung und die Eingabe konnte deshalb nicht gespeichert werden. Soll die Eingabe verworfen werden?"),
                buttons: [
                  DefaultDialogButton(
                    titel: Text("Verwerfen"),
                    onTap: () => dismiss = true,
                  ),
                  DefaultDialogButton(
                    titel: Text("Nein"),
                  )
                ],
              ));
          return dismiss;
        }
      } else {
        saveStateStreamController
            .add(AutoSaveUpdateScaffoldWidget.saveStateNeedSaveInvalid);
        var dismiss = false;
        await showCustomDialog(
            context,
            DefaultDialog(
              dismissible: false,
              title: Text(
                  "Die Eingabe ist fehlerhaft und konnte deshalb nicht gespeichert werden. Korrigiere oder verwerfe die Eingabe."),
              buttons: [
                DefaultDialogButton(
                  titel: Text("Verwerfen"),
                  onTap: () => dismiss = true,
                ),
                DefaultDialogButton(
                  titel: Text("Korrigieren"),
                )
              ],
            ));
        return dismiss;
      }
    } else {
      return true;
    }
  }

  Future<void> _saveItem() async {
    if (mounted)
      saveStateStreamController
          .add(AutoSaveUpdateScaffoldWidget.saveStateSaving);

    if (widget.beforeSave != null) widget.beforeSave!(context, updateItemFrame);

    if (widget.beforeSaveAsync != null)
      await widget.beforeSaveAsync!(context, updateItemFrame);

    if (updateItemFrame!.isNewItem) {
      var itemToSave = updateItemFrame!.editableItem.copy();
      var itemRef =
          await widget.databaseCollection.add(updateItemFrame!.editableItem);
      itemToSave.id = itemRef.id;
      updateItemFrame!.savedItem(itemToSave);
      widget.afterInsert?.call(context, itemRef);
    } else {
      var subItemsChanged =
          (updateItemFrame!.editableItem is HasDatabaseSubItems &&
                  updateItemFrame!.lastSavedItem is HasDatabaseSubItems)
              ? !(updateItemFrame!.editableItem as HasDatabaseSubItems)
                  .equalSubItems(
                      updateItemFrame!.lastSavedItem as HasDatabaseSubItems?)
              : false;
      var updateKeys = updateItemFrame!.getModifiedProperties();
      var itemToSave = updateItemFrame!.editableItem.copy();
      await widget.databaseCollection.update(
        itemToSave,
        onlyUpdateKeys: updateKeys,
        supressSubItemUpdate: !subItemsChanged,
      );
      updateItemFrame!.savedItem(itemToSave);
    }

    if (mounted)
      saveStateStreamController
          .add(AutoSaveUpdateScaffoldWidget.saveStateSaved);
  }

  void doSave(void Function() action) {
    action.call();
    _trySave();
  }

  @override
  void dispose() {
    super.dispose();

    streamSubscriptionInternet?.cancel();
    streamSubscriptionDocument?.cancel();
  }
}

class AutoSaveUpdateItemFrame<T extends UpdateableDatabaseItem<T>> {
  T Function(DocumentSnapshot snapshot)? _factory;

  AutoSaveUpdateItemFrame(this._factory);

  T? _currentDBItem;
  T? _lastSavedItem;
  T? _editableItem;

  void setNewItem(T item) {
    _currentDBItem = item;
    if (_editableItem == null) {
      _lastSavedItem = item.copy();
      _editableItem = item.copy();
    } else {
      var lastSavedItemNew =
          _mergeItemIntoItem(source: _currentDBItem!, target: _lastSavedItem);
      var editableItemNew =
          _mergeItemIntoItem(source: _currentDBItem!, target: _editableItem);
      _lastSavedItem = lastSavedItemNew;
      _editableItem = editableItemNew;
    }
  }

  void savedItem(T savedItem) {
    _lastSavedItem = savedItem.copy();
    _editableItem?.id = savedItem.id;
  }

  bool needSave() {
    return getModifiedProperties().isNotEmpty;
  }

  T _mergeItemIntoItem({required T source, T? target}) {
    var except = getModifiedProperties();
    var newMap = source.toMap().map((key, value) => MapEntry(key,
        except.contains(key) ? target!.toMap()[key] : source.toMap()[key]));
    return _factory!(ReducedDocumentSnapshot(source.id, newMap));
  }

  List<String> getModifiedProperties() {
    T i0 = _lastSavedItem!;
    T i1 = _editableItem!;
    var map0 = i0.toMap();
    var map1 = i1.toMap();

    return map0.keys
        .where((key) =>
            !map1.containsKey(key) ||
            (map1[key]?.toString() ?? "")
                    .compareTo(map0[key]?.toString() ?? "") !=
                0)
        .toList()
          ..addAll(map1.keys.where((key) => !map0.containsKey(key)))
          ..remove("last_modified");
  }

  bool get isNewItem => _editableItem!.id == null;
  T get currentDBItem => _currentDBItem!;
  T get lastSavedItem => _lastSavedItem!;
  T get editableItem => _editableItem!;
}

// ignore: subtype_of_sealed_class
class ReducedDocumentSnapshot implements DocumentSnapshot {
  String? _id;
  Map<String, dynamic> _data;
  ReducedDocumentSnapshot(this._id, this._data);

  @override
  dynamic operator [](field) {
    return _data[field as String];
  }

  @override
  Map<String, dynamic> data() {
    return _data;
  }

  @override
  bool get exists => false;

  @override
  dynamic get(field) {
    return _data[field as String];
  }

  @override
  String get id => _id!;

  @override
  SnapshotMetadata get metadata => throw Exception();

  @override
  DocumentReference get reference => throw Exception();
}

class AutoSaveController {
  void Function() _saveAction;
  AutoSaveController(this._saveAction);

  void save() {
    _saveAction();
  }
}
