import 'dart:math';

import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class DefaultScaffoldWidget extends StatefulWidget {
  final Widget? child;
  final Widget? title;
  final List<Widget>? menuButtons;
  final List<Widget>? appBarButtons;
  final List<Widget>? appBarPopupButtons;
  final WillPopCallback? onWillPop;
  final bool? showBackArrow;
  final bool outsourceMenuButtons;
  final bool titleFullSpace;
  final Key? defaultPageKey;
  final bool scrollable;
  final bool hideAppBar;
  final bool showLoading;
  final IHelpWidget? helpWidget;
  final DefaultScaffoldScrollBarConfiguration? scrollBar;
  final FocusNode? backgroundFocus;

  final List<ScaffoldPageWidget> Function(BuildContext context)?
      updatePagesBuilder;
  final List<ScaffoldPageWidget>? updatePages;

  const DefaultScaffoldWidget({
    Key? key,
    this.child,
    this.title,
    this.menuButtons,
    this.onWillPop,
    this.showBackArrow = false,
    this.appBarButtons = const <Widget>[],
    this.outsourceMenuButtons = false,
    this.updatePagesBuilder,
    this.titleFullSpace = false,
    this.updatePages,
    this.defaultPageKey,
    this.scrollable = true,
    this.hideAppBar = false,
    this.helpWidget,
    this.scrollBar,
    this.showLoading = false,
    this.appBarPopupButtons,
    this.backgroundFocus,
  })  : assert(title != null || hideAppBar),
        assert(child == null || updatePagesBuilder == null),
        assert(updatePages == null || updatePagesBuilder == null),
        assert(updatePages == null || child == null),
        super(key: key);

  @override
  DefaultScaffoldWidgetState createState() => DefaultScaffoldWidgetState();
}

class DefaultScaffoldWidgetState extends State<DefaultScaffoldWidget> {
  Key? currentPageKey;

  ScrollController scrollController = ScrollController();

  late List<ScaffoldPageWidget<dynamic>> updatePages;

  @override
  void initState() {
    super.initState();

    currentPageKey = widget.defaultPageKey;
  }

  void toPage(Key pageKey) {
    setState(() {
      currentPageKey = pageKey;
      updatePages.firstWhere((p) => p.key == pageKey).onShowPage?.call();
    });
  }

  @override
  Widget build(BuildContext context) {
    var draggableDesign = SplitNavigator.mainContext != null &&
        Navigator.of(context) ==
            SplitNavigator.of(SplitNavigatorSide.draggable);

    if ((widget.outsourceMenuButtons) &&
        widget.menuButtons != null &&
        !widget.menuButtons!.every((Widget w) => w.key != null))
      throw ArgumentError(
          "Every menu button of a DefaultScaffoldWidget needs a Key");

    if (widget.outsourceMenuButtons)
      KeyCollection.tabScaffoldKey?.currentState
          ?.showMenuButtons(widget.menuButtons);

    return WillPopScope(
      onWillPop: widget.onWillPop,
      child: Stack(
        children: [
          Container(
            decoration: draggableDesign
                ? BoxDecoration(boxShadow: [
                    BoxShadow(
                      color: Colors.grey.withOpacity(0.5),
                      spreadRadius: 5,
                      blurRadius: 7,
                      offset: Offset(0, 2),
                    )
                  ])
                : null,
            child: Scaffold(
              appBar: widget.hideAppBar
                  ? null
                  : AppBar(
                      actions: getAppBarActions(),
                      backgroundColor:
                          CleanAndroidDesignDefaultColors.background,
                      elevation: 0,
                      automaticallyImplyLeading: false,
                      actionsIconTheme: IconThemeData(
                        size: 27,
                        color: CleanAndroidDesignDefaultColors.black,
                      ),
                      titleSpacing: widget.titleFullSpace
                          ? 0
                          : NavigationToolbar.kMiddleSpacing,
                      title: Row(
                        children: <Widget>[
                          widget.showBackArrow!
                              ? IconButton(
                                  icon: Icon(
                                    Icons.keyboard_arrow_left,
                                    color:
                                        CleanAndroidDesignDefaultColors.black,
                                    size: 30,
                                  ),
                                  onPressed: () async {
                                    if (widget.onWillPop == null ||
                                        await widget.onWillPop!())
                                      Navigator.pop(context);
                                  },
                                )
                              : Container(),
                          SizedBox(
                            width:
                                widget.titleFullSpace && !widget.showBackArrow!
                                    ? 0
                                    : 10,
                          ),
                          Expanded(
                            child: DefaultTextStyle(
                              style: TextStyles.textStyleSize(19).copyWith(
                                  color: CleanAndroidDesignDefaultColors.black),
                              child: widget.title!,
                              overflow: TextOverflow.ellipsis,
                            ),
                          ),
                        ],
                      ),
                    ),
              body: Stack(
                children: <Widget>[
                  Container(
                    decoration: BoxDecoration(
                      color: CleanAndroidDesignDefaultColors.background,
                    ),
                  ),
                  SafeArea(
                    child: GestureDetector(
                      onTap: () {
                        if (widget.backgroundFocus != null)
                          FocusScope.of(context)
                              .requestFocus(widget.backgroundFocus);
                        else
                          FocusScope.of(context).unfocus();
                      },
                      child: Column(
                        children: <Widget>[
                          Expanded(child: _getMainChild()),
                          (widget.outsourceMenuButtons) &&
                                  KeyCollection.tabScaffoldKey?.currentState !=
                                      null
                              ? Container()
                              : Container(
                                  child: Row(
                                    children: widget.menuButtons ?? <Widget>[],
                                  ),
                                  width: MediaQuery.of(context).size.width,
                                  height: widget.menuButtons == null ||
                                          widget.menuButtons!.length == 0
                                      ? 0
                                      : 60,
                                )
                        ],
                      ),
                    ),
                  ),
                  widget.helpWidget != null
                      ? HelpSideWidget(helpWidget: widget.helpWidget!)
                      : Container(),
                ],
              ),
            ),
          ),
          widget.showLoading ? LoadingWidget() : Container(),
        ],
      ),
    );
  }

  List<Widget>? getAppBarActions() {
    if (!(widget.title is Text)) return widget.appBarButtons;

    var list = widget.appBarButtons ?? <Widget>[];
    list = list.where((x) => (!(x is Container) || x.child != null)).toList();

    var titleWidth = (widget.title as Text).getLayoutWidth(
      context,
      explicitTextStyle: TextStyles.textStyleSize(19),
    );
    var screenWidth = MediaQuery.of(context).size.width;
    var spaceForWidgets =
        screenWidth - titleWidth - 45.0 - (widget.showBackArrow! ? 48 : 0);
    var maxWidgets = min(5, (spaceForWidgets / 48.0).floor());

    var newList = <Widget>[];
    var forMenu = <Widget>[];

    if (list.length <= maxWidgets || list.length == 0 || maxWidgets <= 0) {
      newList.addAll(list);
    } else {
      newList.addAll(list.sublist(list.length - maxWidgets + 1));
      forMenu.addAll(list.sublist(0, list.length - maxWidgets + 1));
    }
    forMenu.addAll(widget.appBarPopupButtons ?? <Widget>[]);

    if (forMenu.length > 0)
      newList.add(PopupMenuButton<void Function()>(
        icon: Icon(Icons.more_vert),
        onSelected: (value) => value.call(),
        tooltip: "Mehr zeigen",
        itemBuilder: (context) => forMenu
            .map(
              (item) {
                assert(
                    item is IconButton ||
                        item is PopupMenuEntry<void Function()> ||
                        item is PopupMenuButton<void Function()>,
                    "Only IconButtons can be in the app bar when wrapping in a pop-up menu");

                if (item is PopupMenuEntry<void Function()>) {
                  return item;
                }

                if (item is IconButton) {
                  return PopupMenuItem(
                    value: item.onPressed,
                    child: IconTheme(
                      data: IconThemeData(size: 23),
                      child: Row(
                        children: [
                          item.icon,
                          Container(width: 8),
                          Text(item.tooltip!)
                        ],
                      ),
                    ),
                  );
                }

                if (item is PopupMenuButton<void Function()>) {
                  return PopupMenuItem(
                    value: () {},
                    child: PopupMenuButton<void Function()>(
                      tooltip: item.tooltip,
                      child: Padding(
                        child: IconTheme(
                          data: IconThemeData(size: 23),
                          child: Row(
                            children: [
                              item.icon!,
                              Container(width: 8),
                              Text(item.tooltip!)
                            ],
                          ),
                        ),
                        padding: EdgeInsets.symmetric(vertical: 8.0),
                      ),
                      onSelected: item.onSelected,
                      itemBuilder: item.itemBuilder,
                    ),
                  );
                }

                return PopupMenuItem<void Function()>(
                  child: null,
                );
              },
            )
            .toList()
            .reversed
            .toList(),
      ));
    return newList;
  }

  Widget _getMainChild() {
    Widget? mainChild;
    DefaultScaffoldScrollBarConfiguration? scrollBar = widget.scrollBar;

    if (widget.child != null) {
      mainChild = widget.child;
    } else {
      updatePages = widget.updatePages ?? widget.updatePagesBuilder!(context);

      if (!updatePages.any((p) => p.key == currentPageKey))
        currentPageKey = updatePages.first.key;

      var pageIcons = <ScaffoldPageIcon>[];
      updatePages
          .asMap()
          .forEach((index, page) => pageIcons.add(page.pageIcon.copyWith(
                onTap: () => setState(() {
                  currentPageKey = page.key;
                  updatePages
                      .firstWhere((p) => p.key == page.key)
                      .onShowPage
                      ?.call();
                }),
                isSelected: currentPageKey == page.key,
              )));

      scrollBar =
          updatePages.firstWhere((p) => p.key == currentPageKey).scrollBar;

      mainChild = Column(
        children: <Widget>[
          SizedBox(
            height: pageIcons.length != 0 ? 5 : 0,
          ),
          pageIcons.length != 0
              ? Row(
                  mainAxisAlignment: MainAxisAlignment.spaceEvenly,
                  children: pageIcons,
                )
              : Container(),
          SizedBox(
            height: pageIcons.length != 0 ? 15 : 0,
          ),
          updatePages.firstWhere((p) => p.key == currentPageKey),
        ],
      );
    }

    var hasMenuButtons =
        widget.menuButtons != null && widget.menuButtons!.length > 0;
    return Stack(
      children: [
        widget.scrollable
            ? SingleChildScrollView(
                child: Container(
                  width: MediaQuery.of(context).size.width,
                  child: mainChild,
                ),
                controller: scrollController,
              )
            : mainChild!,
        scrollBar != null
            ? ScrollBarWidget(
                scrollBarConfiguration: scrollBar,
                scrollController:
                    scrollBar.scrollController ?? scrollController,
                cutBottom: hasMenuButtons || widget.outsourceMenuButtons,
              )
            : Container(),
      ],
    );
  }
}
