import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class DefaultListItem extends StatelessWidget implements ISelectableListItem {
  final Widget? leading;
  final IconData? leadingIcon;
  final Widget? trailing;
  final Widget? title;
  final Widget? description;
  final bool? isFirst;
  final bool? isLast;
  final bool? isSelected;
  final GestureTapCallback? onTap;
  final bool Function()? onTapList;
  final void Function()? onLongTapList;
  final void Function(bool)? onHover;
  final HighlightController? highlightController;

  const DefaultListItem({
    Key? key,
    this.leading,
    this.title,
    this.description,
    this.leadingIcon,
    this.trailing,
    this.onTap,
    this.highlightController,
    this.onHover,
  })  : isFirst = false,
        isLast = false,
        isSelected = false,
        onTapList = null,
        onLongTapList = null,
        assert(leading == null || leadingIcon == null),
        super(key: key);

  const DefaultListItem._copy({
    Key? key,
    this.leading,
    this.leadingIcon,
    this.trailing,
    this.title,
    this.description,
    this.isFirst,
    this.isLast,
    this.isSelected,
    this.onTap,
    this.onTapList,
    this.onLongTapList,
    this.highlightController,
    this.onHover,
  }) : super(key: key);

  @override
  Widget build(BuildContext context) {
    var defaultColor = isSelected!
        ? CleanAndroidDesignDefaultColors.selected.withOpacity(0.3)
        : Colors.white;

    if (highlightController == null)
      return buildMainWidget(context, defaultColor);
    else
      return HighlightBuilder(
        controller: highlightController,
        builder: buildMainWidget,
        defaultColor: defaultColor,
      );
  }

  Widget buildMainWidget(BuildContext context, Color color) {
    var borderRadius = BorderRadius.only(
        bottomLeft: isLast! ? Radius.circular(30) : Radius.zero,
        bottomRight: isLast! ? Radius.circular(30) : Radius.zero,
        topLeft: isFirst! ? Radius.circular(30) : Radius.zero,
        topRight: isFirst! ? Radius.circular(30) : Radius.zero);
    return Material(
      borderRadius: borderRadius,
      color: color,
      child: InkWell(
        borderRadius: borderRadius,
        onTap: (onTapList != null || onTap != null)
            ? () {
                if (onTapList == null || onTapList!()) onTap?.call();
              }
            : null,
        onHover: onHover,
        onLongPress: onLongTapList,
        child: Column(
          children: <Widget>[
            Container(
              width: MediaQuery.of(context).size.width,
              decoration: BoxDecoration(
                borderRadius: borderRadius,
              ),
              child: Row(
                children: <Widget>[
                  leading != null || leadingIcon != null
                      ? Padding(
                          padding:
                              EdgeInsets.symmetric(horizontal: 15, vertical: 8),
                          child: Container(
                            width: 45,
                            height: 42,
                            child: Stack(
                              alignment: Alignment.center,
                              children: <Widget>[
                                leading ??
                                    Icon(
                                      leadingIcon,
                                      color: CleanAndroidDesignDefaultColors
                                          .defaultColor,
                                      size: 30,
                                    ),
                                isSelected!
                                    ? Container(
                                        width: 40,
                                        height: 40,
                                        decoration: BoxDecoration(
                                            color:
                                                CleanAndroidDesignDefaultColors
                                                    .selected
                                                    .withOpacity(0.7),
                                            borderRadius:
                                                BorderRadius.circular(20)),
                                        child: Icon(
                                          Icons.check,
                                          color: Colors.white,
                                        ),
                                      )
                                    : Container(),
                              ],
                            ),
                          ),
                        )
                      : Container(
                          width: 20,
                        ),
                  Expanded(
                    child: Column(
                      crossAxisAlignment: CrossAxisAlignment.start,
                      children: <Widget>[
                        title != null
                            ? DefaultTextStyle(
                                style: TextStyles.textStyleSize(17).copyWith(
                                    color:
                                        CleanAndroidDesignDefaultColors.black),
                                child: title!)
                            : Container(),
                        description != null
                            ? DefaultTextStyle(
                                style: TextStyles.textStyleSize(13),
                                child: description!)
                            : Container(),
                      ],
                    ),
                  ),
                  Padding(
                    padding:
                        const EdgeInsets.symmetric(horizontal: 15, vertical: 0),
                    child: trailing ?? Container(),
                  ),
                ],
              ),
            ),
            !isLast!
                ? Padding(
                    padding: const EdgeInsets.symmetric(horizontal: 20),
                    child: Divider(
                      color: CleanAndroidDesignDefaultColors.defaultColorLight,
                      height: 1.1,
                      thickness: 1.1,
                    ),
                  )
                : Container()
          ],
        ),
      ),
    );
  }

  IListItem copyWithSelection(
      {bool? isSelected,
      bool Function()? onTapList,
      void Function()? onLongTapList}) {
    return DefaultListItem._copy(
      key: this.key,
      description: this.description,
      leading: this.leading,
      leadingIcon: this.leadingIcon,
      title: this.title,
      isFirst: this.isFirst,
      isLast: this.isLast,
      onTap: this.onTap,
      onHover: this.onHover,
      trailing: this.trailing,
      highlightController: this.highlightController,
      isSelected: isSelected ?? this.isSelected ?? false,
      onTapList: onTapList ?? this.onTapList,
      onLongTapList: onLongTapList ?? this.onLongTapList,
    );
  }

  IListItem copyWithPosition({bool? isFirst, bool? isLast}) {
    return DefaultListItem._copy(
      key: this.key,
      description: this.description,
      leading: this.leading,
      leadingIcon: this.leadingIcon,
      title: this.title,
      onTap: this.onTap,
      onHover: this.onHover,
      isSelected: this.isSelected,
      onTapList: this.onTapList,
      onLongTapList: this.onLongTapList,
      trailing: this.trailing,
      highlightController: this.highlightController,
      isFirst: isFirst ?? this.isFirst,
      isLast: isLast ?? this.isLast,
    );
  }
}

abstract class ISelectableListItem extends IListItem {
  // ignore: missing_return
  IListItem copyWithSelection(
      {bool? isSelected,
      bool Function()? onTapList,
      void Function()? onLongTapList}) {
    throw Exception();
  }

  @override
  // ignore: missing_return
  Element createElement() {
    throw Exception();
  }
}
