import 'package:clean_android_design/clean_android_design.dart';
import 'package:flutter/material.dart';

class NavigatorWidget extends StatefulWidget {
  final Widget? menuWidget;

  const NavigatorWidget({
    Key? key,
    this.menuWidget,
  }) : super(key: key);

  @override
  NavigatorWidgetState createState() => NavigatorWidgetState();
}

class NavigatorWidgetState extends State<NavigatorWidget> {
  final RouteObserver<PageRoute> routeObserver = RouteObserver<PageRoute>();
  final RouteObserver<PageRoute> draggableRouteObserver =
      RouteObserver<PageRoute>();

  bool showRightSide = false;
  bool showDraggable = false;

  double? draggableWidth, draggableHeight;
  Offset? draggablePosition;

  @override
  Widget build(BuildContext context) {
    SplitNavigator.mainContext = context;
    SplitNavigator.navigatorWidgetState = this;

    return Stack(
      children: [
        _buildMainNavigators(),
        _buildDraggable(),
      ],
    );
  }

  Widget _buildMainNavigators() {
    return Scaffold(
      body: Row(
        children: <Widget>[
          widget.menuWidget ?? Container(),
          Expanded(
            child: LockedNavigatorWidget(
              child: Navigator(
                key: SplitNavigator.leftNavigatorKey,
                onGenerateRoute: (settings) => MaterialPageRoute(
                  builder: (context) {
                    return Container();
                  },
                ),
              ),
            ),
          ),
          Container(
            width: 20,
            height: MediaQuery.of(context).size.height,
            color: CleanAndroidDesignDefaultColors.background,
          ),
          AnimatedContainer(
            duration: Duration(milliseconds: 150),
            width: showRightSide ? (MediaQuery.of(context).size.width / 2) : 0,
            child: Navigator(
              key: SplitNavigator.rightNavigatorKey,
              onGenerateRoute: (settings) =>
                  NoPageRoute(routeObserver, setShowRightSide, LoadingWidget()),
              observers: [routeObserver],
            ),
          ),
        ],
      ),
    );
  }

  Widget _buildDraggable() {
    if (draggablePosition == null) {
      draggableWidth = MediaQuery.of(context).size.width / 2;
      draggableHeight = MediaQuery.of(context).size.height * 0.8;
      draggablePosition = Offset(MediaQuery.of(context).size.width * 0.2,
          MediaQuery.of(context).size.height * 0.1);
    }

    var draggableChild = AnimatedContainer(
      duration: Duration(milliseconds: 300),
      width: draggableWidth,
      height: showDraggable ? draggableHeight : 0,
      child: Navigator(
        key: SplitNavigator.draggableNavigatorKey,
        onGenerateRoute: (settings) =>
            NoPageRoute(draggableRouteObserver, setShowDraggable, Container()),
        observers: [draggableRouteObserver],
      ),
    );

    return Positioned(
      left: draggablePosition!.dx,
      top: draggablePosition!.dy,
      child: Draggable(
        child: draggableChild,
        childWhenDragging: Container(),
        feedback: draggableChild,
        onDraggableCanceled: (Velocity velocity, Offset offset) {
          setState(() => draggablePosition = offset);
        },
      ),
    );
  }

  void setShowRightSide(bool showRightSide) {
    Future.delayed(
        Duration.zero,
        () => setState(() {
              this.showRightSide = showRightSide;
              Future.delayed(Duration.zero, () {
                SplitNavigator.rightIsShown = showRightSide;
                SplitNavigator.updateKey.update();
              });
            }));
  }

  void setShowDraggable(bool showDraggable) {
    Future.delayed(
        Duration.zero,
        () => setState(() {
              this.showDraggable = showDraggable;
              Future.delayed(Duration.zero, () {
                SplitNavigator.draggableIsShown = showDraggable;
                SplitNavigator.updateKey.update();
              });
            }));
  }

  void setDraggablePositionAndSize({
    Offset? position,
    double? height,
    bool heightAbsoluteValue = true,
    double? width,
    bool widthAbsoluteValue = true,
  }) {
    setState(() {
      if (position != null) draggablePosition = position;
      if (height != null)
        draggableHeight = height *
            (heightAbsoluteValue ? 1 : MediaQuery.of(context).size.height);
      if (width != null)
        draggableWidth = width *
            (widthAbsoluteValue ? 1 : MediaQuery.of(context).size.width);
    });
  }
}

class NavigatorListener with RouteAware {
  int n = -1;

  final void Function(bool showNavigator) onChange;

  NavigatorListener(
    this.onChange,
  );

  @override
  void didPush() {
    super.didPush();
    n++;
    if (n == 1) onChange(true);
  }

  @override
  void didPushNext() {
    super.didPushNext();
    n++;
    if (n == 1) onChange(true);
  }

  @override
  void didPop() {
    super.didPop();
    n--;
    if (n == 0) onChange(false);
  }

  @override
  void didPopNext() {
    super.didPopNext();
    n--;
    if (n == 0) onChange(false);
  }
}

class NoPageRoute extends MaterialPageRoute {
  NoPageRoute(
    RouteObserver<PageRoute> routeObserver,
    void Function(bool showNavigator) onChange,
    Widget noChild,
  ) : super(
          builder: (context) {
            routeObserver.subscribe(
                NavigatorListener(onChange), ModalRoute.of(context) as PageRoute<dynamic>);
            return Center(
              child: noChild,
            );
          },
        );
}
